#include "ObjectList.h"
#include "MemUtils.h"

#define kGrowIncrement	4

//------------------------------------------------------------------------------

DefineClassNoBase(ObjectList);

//------------------------------------------------------------------------------

ObjectList::~ObjectList()
{
	if (fObjects != nil)
	{
		DeleteAll();
	}
}

//------------------------------------------------------------------------------

long ObjectList::CountItems() const
{
	return fCount;
}

//------------------------------------------------------------------------------

long ObjectList::GetObjectIndex(TObject* obj) const
{
	long count = fCount;
	long index = 0;

	if (count > 0)
	{
		ObjectListPtr p = *fObjects;
		
		while (index < count)
		{
			if (p->objects[index++] == obj)
			{
				return index;
			}
		} 
	}
	
	return 0;
}

//------------------------------------------------------------------------------

SmartPtr<TObject> ObjectList::GetNthObject(
	long			index,
	CvtOptions		/*access*/)	const
{
	if (index < 0) // convert to zero based
	{
		index += fCount;
	}
	else
	{
		index -= 1;
	}

	if (index < 0 || index >= fCount)
	{
		return nil;
	}
	else
	{
		return (**fObjects).objects[index];
	}
}

//------------------------------------------------------------------------------

void ObjectList::AppendObject(TObject* obj, bool once)
{
	if (!once || !obj || !GetObjectIndex(obj))
	{
		fObjects = InsertObjects(fObjects, 0, 1, &obj);
		fCount  += 1;
	}
}

//------------------------------------------------------------------------------

void ObjectList::PrependObject(TObject* obj, bool once)
{
	if (!once || !obj || !GetObjectIndex(obj))
	{
		fObjects = InsertObjects(fObjects, 1, 1, &obj);
		fCount  += 1;
	}
}

//------------------------------------------------------------------------------

void ObjectList::InsertObjectAt(long index, TObject* obj)
{
	fObjects = InsertObjects(fObjects, index, 1, &obj);
	fCount  += 1;
}

//------------------------------------------------------------------------------

void ObjectList::ReplaceObjectAt(long index, TObject* obj)
{
	if (index < 0) // convert to zero based
	{
		index += fCount + 1;
	}
	else
	{
		index -= 1;
	}
	
	if (index >= 0 && index < fCount)
	{
		ObjectPtr	ref(obj); // bump our refrence count
		TObject*&	sav = (**fObjects).objects[index];
		TObject*	tmp = sav;
		
		sav = ref.ptr;	// move to our array
		ref.ptr = tmp;	// ref will dispose of old object on exit
	}
	else if (index == -1 || index == fCount)
	{
		AppendObject(obj);
	}
	else
	{
		Debugger();
	}	
}

//------------------------------------------------------------------------------

void ObjectList::DeleteAll()
{
	long count = fCount;
	
	fCount = 0;				// steal the count	

	if (fObjects)			
	{
		ObjectListHdl objects = fObjects;

		CTempHandle h((Handle)objects);

		fObjects = nil;
		
		ReleaseObjects(objects);
	}
}

//------------------------------------------------------------------------------

void ObjectList::DeleteAt(long index)
{
	long count = fCount;
	
	if (index < 0)
		index += count;
	else
		index -= 1;
		
	if (index >= 0 && index < count)
	{
		if (count == 1)							// Last element
		{
			DeleteAll();
		}
		else
		{
			ObjectListPtr p = *fObjects;
			
			fCount = count -= 1;				// update the count
			
			TObject* obj = p->objects[index];	// cache the object
			
			if (index < count)					// collapse it out 
			{
				BlockMoveData(&p->objects[index + 1],
							  &p->objects[index],
							  (count-index)*sizeof(TObject*));
			}
		
			if (p->count > count + kGrowIncrement)
			{
				p->count = count;				// Trim handle if necessary
				SetHandleSize((Handle) fObjects,
							  sizeof(ObjectList) - sizeof(TObject*) + count*sizeof(TObject*));
			}
			else							
			{
				p->objects[count] = nil;		// otherwise clear the free space
			}
			
			if (obj != nil)
			{
				obj->DeleteReference();
			}
		}
	}
}

//------------------------------------------------------------------------------

void ObjectList::DeleteRange(long start, long end)
{
	long count = fCount;

	if (count > 0)
	{
		if (start < 0)		// convert to positive, one based index
			start += count + 1;
			
		if (end < 0)		// convert to positive, one based index
			end += count + 1;
			
		if (start <= 0)
		{
			start = 1;
		}

		if (end > count)
		{
			end = count;
		}
		
		if (end == start)		// special case: single element
		{
			DeleteAt(start);
		}
		else
		{
			long num = end - (--start);	// got to zero-base	

			if (num >= count)			// special case: all of it
			{
				DeleteAll();
			}
			else
			{
				ObjectListHdl kill = ExtractObjects(fObjects, start, num);
				
				CTempHandle	h((Handle) kill);
				ReleaseObjects(kill);
			}
		}
	}
}

//------------------------------------------------------------------------------

void ObjectList::DeleteObject(TObject* obj, bool once)
{
	long index;
	
	do
	{
		index = GetObjectIndex(obj);
		
		if (index != 0)
		{
			DeleteAt(index);
		}
	}
	while(index && !once);
}

//------------------------------------------------------------------------------

void ObjectList::DeleteAllBut(long start, long end)
{
	long count = fCount;

	if (count > 0)
	{
		if (start < 0)		// convert to positive, zero based index
			start += count;
		else
			start--;
			
		if (end < 0)		// convert to positive, one based index
			end += count + 1;
			
		if (start < 0)
		{
			start = 0;
		}

		if (end > count)
		{
			end = count;
		}

		if (start > 0 || end < count)
		{
			long num = end - start;
			
			if (num <= 0)
			{
				DeleteAll();
			}
			else
			{
				ObjectListHdl kill = fObjects;
				
				fObjects = ExtractObjects(kill, start, num);
				
				CTempHandle	h((Handle) kill);
				ReleaseObjects(kill);
			}
		}
	}	
}

//------------------------------------------------------------------------------

ObjectList::ObjectListHdl ObjectList::AllocObjectList(
	long			count,
	ObjectListHdl	h,
	bool			maintain)
{
	Size size = count * sizeof(TObject*) + sizeof(ObjectList) - sizeof(TObject*);

	if (h != nil)
	{
		long oldcount = (**h).count;
		
		long delta = count - oldcount; // Number of cells we're adding (removing)
		
		if (delta != 0)
		{
			if (delta < 0)	// We're shrinking, realease any old objects
			{
				if (maintain) 
				{
					ReleaseObjects(h, -delta, count);
				}
			}
		
			SetHandleSize((Handle) h, size);
			FailMemError();

			if (delta > 0)	// We're growing, zero the new space
			{
				memset(&(**h).objects[oldcount], 0, delta * sizeof(TObject*));
			}
		}
	}
	else
	{
		FailNIL(h = (ObjectListHdl) NewHandleClear(size));
	}
	
	(**h).count = count;
	
	return h;
}

//------------------------------------------------------------------------------

ObjectList::ObjectListHdl ObjectList::ExtractObjects(
	ObjectListHdl	src,
	long			index,
	long			count,
	ObjectListHdl	dst)
{
	long end = src ? (**src).count : 0;

	if (index >= end)
	{
		index = end;
		count = 0;
	}
	else if (count + index > end)
	{
		count = end - index;
	}
	
	dst = AllocObjectList(count, dst, true); // (re)allocate the destination

	if (dst != nil && count > 0)
	{
		ObjectListPtr psrc = *src;
		ObjectListPtr pdst = *dst;

		long tail = end - (index + count);
		
		BlockMoveData(psrc->objects[index],
					  pdst->objects[0],
					  count * sizeof(TObject*));
		
		if (tail > 0) // squeeze down anything that's left
		{
			BlockMoveData(psrc->objects[index+count],
						  psrc->objects[index],
						  tail * sizeof(TObject*));
			
			index += tail;
		}
		
		AllocObjectList(index, src, false); // resize the source
	}

	return dst;
}

//------------------------------------------------------------------------------

ObjectList::ObjectListHdl ObjectList::ReleaseObjects(
	ObjectListHdl	h,
	long			count,
	long			offset)
{
	long end	= (**h).count;
	long index	= offset;

	if (count < 0)
	{
		count = (**h).count - offset;
	}

	while (count-- > 0 && index < end)	// kill off the objects
	{
		try
		{
			TObject** p   = &(**h).objects[index++];
			TObject*  obj = *p;
			
			*p = nil;
			
			if (obj != nil)
			{
				obj->DeleteReference();
			}
		}
		catch(...)
		{
		}
	}
	
	return h;
}

//------------------------------------------------------------------------------

ObjectList::ObjectListHdl ObjectList::InsertObjects(
	ObjectListHdl	list,
	long			index,
	long			count,
	TObject**		data)
{
	long oldcount = list ? (**list).count : 0;

	if (index <= 0)		// convert to positive zero-base
	{
		index += oldcount;
	}
	else
	{
		index -= 1;
	}
	
	if (index < 0 || index > oldcount)
	{
		Debugger();		// range check
		
		index = index < 0 ? 0 : oldcount;
	}
	
	list = AllocObjectList(oldcount + count, list, true);

	if (data != nil)	// copy in the objects
	{
		for (int i = 0; i < count; i++)
		{
			TObject* obj = data[i];
			
			if (obj)
			{
				obj->NewReference();
			}
			
			(**list).objects[index++] = obj;
		}
	}

	return list;
}

//==============================================================================

DefineClass2Base(TObjectList, TObject, ObjectList);

//------------------------------------------------------------------------------

TObjectList::~TObjectList()
{
}

//------------------------------------------------------------------------------

void TObjectList::Finalize()
{
	DeleteAll();
	
	Inherited::Finalize();
}

//------------------------------------------------------------------------------

